/////////////////////////////////////////////////////////////////////
//Beispiel für einen Filter, der auf der Fouriertransformation beruht
//aus Kapitel BildbearbeitungS
////////////////////////////////////////////////////////////////////

int     bufferSize=512;
float[][] ir, ii, or, oi;

float modifier1=0; // untere Filterfrequenz
float modifier2=384; // obere Filterfrequenz
int mapping = 1;

PImage inp, powerSpektrum, gefiltertesBild; 

//////////////////////////////////////////////////

void generiereFunktionUndFFT() {

  for (int i=0; i<bufferSize; i++) 
    for (int k=0; k<bufferSize; k++) {
      ir[i][k] = red(inp.get(i, k));
      ii[i][k] = 0;
    }

  for (int i=0; i<bufferSize; i++) 
    for (int k=0; k<bufferSize; k++) {
      or[i][k] = ir[i][k];
      oi[i][k] = ii[i][k];
    } 

  fft_2D(1, bufferSize, or, oi);

  // wandele Array so um, dass tiefste Frequenzen in der
  // Mitte sind
  zentriereFrequenzen(or, oi); 
  // schneide Frequenzen heraus
  filtereFrequenzen(or, oi, modifier1, modifier2); 
  // zeige das gefilterte Powerspektrum
  powerSpektrum = erzeugeFFTausgabe(or, oi);   
  // wandele Frequenzarray wieder zurück
  dezentriereFrequenzen(or, oi);

  fft_2D(-1, bufferSize, or, oi);

  for (int i=0; i<bufferSize; i++) 
    for (int k=0; k<bufferSize; k++) {
      float val = or[i][k];
      gefiltertesBild.set(i, k, color(val, val, val));
    }
}

////////////////////////////////////////////////////////
// 
// nimm nur Frequenzen, die zwischen f1 und f2
// liegen, alles andere wird null
//
///////////////////////////////////////////////////////

void filtereFrequenzen(float[][] vr, float[][] vi, float f1, float f2) {

  for (int i=0; i<bufferSize; i++) 
    for (int k=0; k<bufferSize; k++) {
      float dist = sqrt(sq(i-bufferSize/2)+sq(k-bufferSize/2));
      if ((dist<f1) || (dist>f2)) {
        vr[i][k] = 0;
        vi[i][k] = 0;
      }
    }
}

//////////////////////////////////////////////////
//  
// zeige die Ergebnisse der Fouriertransformierte
// in logarithmisierter Form und auf zentrierte 
// Weise mit tiefen Frequenzen in der Mitte
//
//////////////////////////////////////////////////

float mapValue(float r, float i, float minVal, float maxVal) {
  float val = 0;

  switch (mapping) {
  case 1: 
    val = 255*log(sqrt(sq(r)+sq(i)) - minVal)/log(maxVal-minVal); 
    break;
  case 2: 
    val = 255*(sqrt(sq(r)+sq(i))-minVal)/(maxVal-minVal); 
    break;
  }

  return val;
}

////////////////////////////////////////////////////////

PImage erzeugeFFTausgabe(float[][] or, float[][] oi) {

  PImage outp = createImage(or.length, or.length, RGB);
  int bs = or.length;
  int bs2 = or.length/2;

  float maxValue = 0;
  float minValue=1e6;
  for (int i=0; i<bs; i++) 
    for (int k=0; k<bs; k++) {
      maxValue = max(maxValue, sqrt(sq(or[i][k])+sq(oi[i][k])));
      minValue = min(minValue, sqrt(sq(or[i][k])+sq(oi[i][k])));
    }

  for (int i=0; i<bs; i++) 
    for (int k=0; k<bs; k++) {
      float value = mapValue(or[i][k], oi[i][k], minValue, maxValue);
      outp.set(i, k, color(value, value, value));
    }

  return outp;
}

//////////////////////////////////////////////////

void setup()
{
  size(1600, 512, JAVA2D);
  noLoop();

  ir = new float[bufferSize][bufferSize];
  ii = new float[bufferSize][bufferSize];
  or = new float[bufferSize][bufferSize];
  oi = new float[bufferSize][bufferSize];

  inp = loadImage("DFT_Vorlage.jpg");
  inp.resize(bufferSize, bufferSize);
  powerSpektrum = createImage(bufferSize, bufferSize, RGB);
  gefiltertesBild = createImage(bufferSize, bufferSize, RGB);

  generiereFunktionUndFFT();
}

//////////////////////////////////////////////////

void draw()
{
  background(255);
  image(inp, 0, 0);
  image(powerSpektrum, width/2-bufferSize/2, 0);
  image(gefiltertesBild, width-bufferSize, 0);
}

//////////////////////////////////////////////////

void keyPressed() {

  switch (key) {
  case 'y' : mapping=1; // logarithmisches mapping
             break; 
  case 'x' : mapping=2; // lineares mapping  
             break; 
  case '+' : modifier2 += 4; // obere Frequenz
             modifier2 = min(1.5*bufferSize/2,modifier2);
             break; 
  case '-' : modifier2 -= 4; 
             modifier2 = max(0,modifier2);
             break;
  case 'q' : modifier1 += 4; // untere Frequenz
             modifier1 = min(1.5*bufferSize/2,modifier1);
             break; 
  case 'w' : modifier1 -= 4; 
             modifier1 = max(0,modifier1);
             break;
  case 's' : save("test.png"); 
             break;
  }
  println("Untere Frequenz: " + modifier1 + " obere Frequenz: " + modifier2);

  generiereFunktionUndFFT();
  redraw();
}            